'''
..
  $Id: proxy.py 97 2018-08-23 12:53:55Z pytools $

  OSGi python tools of the clazzes.org project
  http://www.clazzes.org

  Created: 13.04.2020

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import traceback
import unittest
import logging
import threading
import os
import datetime
from clazzes.http import xmlrpc
from clazzes.osgitools.utils import retryOperation
from paramiko.rsakey import RSAKey
import subprocess
from clazzes.osgitools.ssh import OSGiClient
from clazzes.osgitools.testrail import TestrailServer, TestrailResult, STATUS_Failed, STATUS_Passed, TestrailStepResult, TestrailCase, TestrailStep
from requests import request

log = logging.getLogger(__name__)

_THREAD_LOCAL = threading.local()

def getCurrentUser():
    '''
       Factory method for LoginInfo objects, which are bound through the
       CurrentUser context manager.
    '''
    return getattr(_THREAD_LOCAL,"currentUser",None)

class CurrentUser:
    '''
       A context manager, which sets the thread-bound current user
       to be returned by ~getCurrentUser
    '''
    def __init__(self, loginInfo : xmlrpc.LoginInfo):
        '''
           Initialize the context manager with a given ~xmlrpc.LoginInfo
           user information.
        '''
        self.loginInfo = loginInfo
        self.oldLoginInfo = None

    def __enter__(self):
        self.oldLoginInfo = getCurrentUser()
        _THREAD_LOCAL.currentUser = self.loginInfo
        return self

    @property
    def user(self):
        '''
           Return the user bound to the current thread by this context manager.
        '''
        return _THREAD_LOCAL.currentUser

    def __exit__(self,exc_type,exc_val,exc_tb):  # @UnusedVariable
        _THREAD_LOCAL.currentUser = self.oldLoginInfo

def getCurrentCertificate():
    '''
       Factory method for extra headers dicts, which are bound through the
       CurrentCertificate context manager.
    '''
    return getattr(_THREAD_LOCAL,"currentCertificate",None)

def buildCertificate(subjectDn : str, serial: int, issuerDn : str):
    '''
       Build certificate headers for use with the CurrentCertificate context manager.

       :param subjectDn: The subject DN of the certificate to pass.
       :param serial: The subject DN of the certificate to pass as HEX-formatted header.
       :param issuerDn: The issuer DN of the certificate to pass.
    '''
    return {
        "SSL_CLIENT_S_DN": subjectDn,
        "SSL_CLIENT_M_SERIAL": f"{serial:X}",
        "SSL_CLIENT_I_DN": issuerDn
    }

class CurrentCertificate:
    '''
       A context manager, which sets the thread-bound current certificate
       headers to be returned by ~getCurrentCertificate
    '''
    def __init__(self, headers : dict):
        '''
           Initialize the context manager with a given ~xmlrpc.LoginInfo
           user information.
        '''
        self.headers = headers
        self.oldHeaders = None

    def __enter__(self):
        self.oldHeaders = getCurrentCertificate()
        _THREAD_LOCAL.currentCertificate = self.headers
        return self

    @property
    def certificate(self):
        '''
           Return the headers bound to the current thread by this context manager.
        '''
        return _THREAD_LOCAL.currentCertificate

    def __exit__(self,exc_type,exc_val,exc_tb):  # @UnusedVariable
        _THREAD_LOCAL.currentCertificate = self.oldHeaders


class OSGiTest(unittest.TestCase):
    '''
       A test case, which controls an OSGi Runner instance via SSH and HTTP.
       The test is configured by overriding class variables.
    '''

    baseUrl = "http://localhost:8082/"
    '''
       The base URL
    '''
    sshPort = 8102
    '''
       The ssh port to connect to.
    '''
    sshPrivateKeyFile = None
    '''
       The path of private key file to read.
    '''
    sshPrivateKey= None
    '''
       The parsed private key initialized by ~setUpClass.
    '''
    sshKnownHostsFile = None
    '''
       The path of the known_hosts file to read.
    '''
    sshRetries = 30
    '''
       The number of retries when connecting to the SSH console.
    '''
    dataSourceName = None
    '''
       The name of the JDBC datasource to operate on.
    '''
    dataSourceSchema = None
    '''
       The optional schema of database tables. This must be set to `public` for postgresql.
    '''
    osgiDir = None
    '''
       The path of the osgi runner test setup used to start the osgi-runner instance.
    '''
    osgiClient = None
    '''
      The ~OSGiClient instance initialized by ~setUpClass.
    '''

    doRun = False
    '''
       If True, start the osgi-runner instance in ~osgiDir.
    '''
    doDockerCompose = False
    '''
       If True, start `docker-compose -d` in ~osgiDir.
    '''
    dockerComposeLogFile = None
    '''
       If set, save the docker-compose logs to the given path before stopping.
       stdout is postfixed by `-stdout.log` , stderr is postfixed by `-stderr.log`
    '''
    dockerComposeFiles = None
    '''
       If you need to load more than one docker-compose files apart from `docker-compose.yml`,
       set this variable to a sequence of file names.
    '''
    runArgs = None
    '''
       Optional additional arguments to pass to the osgi-runner instance in ~osgiDir or to
       `docker-compose -d`.
    '''
    osgiRunner = None
    '''
       Popen object of running osgi-runner instance.
    '''
    loginMechanism = 'org.clazzes.login.jaas'
    '''
       The configured login mechanism.
    '''
    loginService = None
    '''
       The login service set up by ~setUpClass.
    '''

    testrailServer = None
    '''
       The testrail-server to push test results to or none to disable testrail support.
    '''
    testrailRunId = None
    '''
       The testrail run ID for to store test results to.
    '''
    applicationVersion = None
    '''
       The application version under test, if known.
    '''

    @classmethod
    def setUpClass(cls):

        super(OSGiTest,cls).setUpClass()

        # fetch SSH private key
        if cls.sshPrivateKeyFile:
            cls.sshPrivateKey = RSAKey.from_private_key_file(cls.sshPrivateKeyFile)

        # FIXME start osgi-runner
        if cls.doRun and cls.osgiDir:

            popenArgs = [os.path.join(cls.osgiDir,"run.sh")]

            if cls.runArgs:

                if isinstance(cls.runArgs,str):
                    popenArgs.append(cls.runArgs)
                else:
                    popenArgs.extend(cls.runArgs)

            cls.osgiRunner = subprocess.Popen(popenArgs)

        if cls.doDockerCompose and cls.osgiDir:

            composeArgs = ["docker-compose"]

            if cls.dockerComposeFiles:
                for composeFile in cls.dockerComposeFiles:
                    composeArgs.append("-f")
                    composeArgs.append(composeFile)

            composeArgs.extend(["up","-d"])

            if cls.runArgs:

                if isinstance(cls.runArgs,str):
                    composeArgs.append(cls.runArgs)
                else:
                    composeArgs.extend(cls.runArgs)

            log.info("Running %s"%(composeArgs,))
            subprocess.run(composeArgs,cwd=cls.osgiDir)

        try:

            cls.osgiClient = OSGiClient(cls.sshPort,cls.sshPrivateKey,cls.sshKnownHostsFile,cls.dataSourceName,cls.dataSourceSchema)

            retryOperation("OSGiClient.connect",cls.osgiClient.connect,ntries=cls.sshRetries)

            cls.setupServices()

        except:

            log.exception("Caught exception in setUpClass()")

            if cls.osgiClient:
                log.info("Closing OSGiClient upon exception.")
                cls.osgiClient.close()
                cls.osgiClient = None
            if cls.osgiRunner:
                log.info("Stopping osgi-runner upon exception.")
                cls.osgiRunner.terminate()
                log.info("Waiting for osgi-runner to stop upon exception.")
                rc = retryOperation("stop osgi-runner",cls.osgiRunner.wait,(1,),ntries=20,interval=0.1)
                log.info("osgi-runner stopped with return code [%s] upon exception."%rc)
                #cls.osgiRunner.kill()
                #cls.osgiRunner.communicate()
                cls.osgiRunner = None

            if cls.doDockerCompose and cls.osgiDir:
                log.info("Running `docker-compose down`.")
                subprocess.run(("docker-compose","down"),cwd=cls.osgiDir,check=True)

            raise

    @classmethod
    def tearDownClass(cls):
        super(OSGiTest, cls).tearDownClass()
        if cls.osgiClient:
            log.info("Closing OSGiClient upon tear down.")
            cls.osgiClient.close()
            cls.osgiClient = None
        if cls.osgiRunner:
            log.info("Stopping osgi-runner with PID %d upon tear down." % cls.osgiRunner.pid)
            cls.osgiRunner.terminate()
            log.info("Waiting for osgi-runner to stop upon tear down.")
            rc = retryOperation("stop osgi-runner",cls.osgiRunner.wait,(1,),ntries=20,interval=0.1)
            log.info("osgi-runner stopped with return code [%s] upon tear down."%rc)
            #log.info("osgiRunner.kill() ...")
            #cls.osgiRunner.kill()
            #log.info("osgiRunner.communicate() ...")
            #cls.osgiRunner.communicate()
            #log.info("None-ing osgiRunner")
            cls.osgiRunner = None
            log.info("Done tearing down osgi-runner")
        if cls.doDockerCompose and cls.osgiDir:
            if cls.dockerComposeLogFile:
                r = subprocess.run(("docker-compose","logs"),cwd=cls.osgiDir,capture_output=True)
                fn = cls.dockerComposeLogFile+"-stdout.log"
                log.info("Saving docker-compose stdout to [%s]."%fn)
                with open(fn,mode="wb") as fh:
                    fh.write(r.stdout)
                fn = cls.dockerComposeLogFile+"-stderr.log"
                log.info("Saving docker-compose stderr to [%s]."%fn)
                with open(fn,mode="wb") as fh:
                    fh.write(r.stderr)

            log.info("Running `docker-compose down`.")
            subprocess.run(("docker-compose","down"),cwd=cls.osgiDir)


    @classmethod
    def createService(cls,name,url,needLogin = True):
        '''
           Helper class method to set up an XMLRPC service and a likewise service bean.

           :param name: The name of the class attribute to which the ~xmlrpc.RPCService for
                        the given URL will be bound. The service bean will be bound to then
                        ``name+"Bean"`` class attribute.
           :param url: The URL relative to ~baseUrl of the service.
           :param needLogin: If True, the service bean will have ~getCurrentUser
                             as login_info_factory.
        '''
        svc = xmlrpc.RPCService(cls.baseUrl,url)
        setattr(cls,name,svc)

        smd = retryOperation(name+" SMD",svc.retrieveSMDInfo)
        svcBean = xmlrpc.makeBean(svc,smd,login_info_factory=getCurrentUser if needLogin else None)
        setattr(cls,name+"Bean",svcBean)

    @classmethod
    def setupServices(cls):
        '''
           A hook function called by ~OSGiTest.setUpClass, which may be overwritten by
           subclasses.

           The base implementation sets up the login service with the given mechanism.
        '''
        cls.loginService = xmlrpc.LoginService(cls.baseUrl,cls.loginMechanism)

    @classmethod
    def trCase(cls, caseId: int, reentrant: bool = False) -> TestrailCase:
        '''
           Return a context manager for a testrail test case.
           You should call passed() on the returned instance to commit test success to testrail.
           Any exceptions thrown within the context handler will propagate to the TR instance as "Failure".

        :param caseId: The testaril test case id.
        :type caseId: int
        :param reentrant: If this test can be safely repeated without reloading the data base. Typically
            'False' for any cases which demand DB writes.
        :return: _description_
        :rtype: TestrailCase
        '''
        return TestrailCase(cls.testrailServer,cls.testrailRunId,caseId,cls.applicationVersion,reentrant=reentrant)

    @classmethod
    def trStep(cls, content: str, expected: str, tc: TestrailCase) -> TestrailStep:
        '''
            Return a context manager for a testrail step.
            Use passed() within this context handler to add the passed step to the TestrailResult of this case.
            Any exceptions thrown within the context handler will add a failed step to the TestrailResult and
            also fail the test immediately (any remaining steps are skipped)
        '''

        return TestrailStep(content, expected, tc)

    def assertNeedsLogin(self,func,*args,**kwargs):
        '''
           Checks, whether the given bean call need a login, aka it throws
           an ~xmlrpc.XmlRpcTargetError with error code ~xmlrpc.JSON_RPC_LOGIN_REQUIRED
        '''
        with self.assertRaises(xmlrpc.XmlRpcTargetError) as cm:
            func(*args,**kwargs)

        self.assertEqual(xmlrpc.JSON_RPC_LOGIN_REQUIRED,cm.exception.code)
        log.info("[OK] Function [%s] requires a login."%(xmlrpc.methodName(func),))

    def assertUnauthorized(self,func,*args,**kwargs):
        '''
           Checks, whether the given bean call need a login, aka it throws
           an ~xmlrpc.XmlRpcTargetError with error code ~xmlrpc.JSON_RPC_SERVER_ERROR_0-1
        '''
        with self.assertRaises(xmlrpc.XmlRpcTargetError) as cm:
            func(*args,**kwargs)

        self.assertEqual(xmlrpc.JSON_RPC_SERVER_ERROR_0-1,cm.exception.code)
        log.info("[OK] Function [%s] is unauthorized."%(xmlrpc.methodName(func),))

    def assertAllNeedLogin(self,service):
        '''
           Checks, whether all bean calls of the given service bean need a login
           by invoking ~assertNeedsLogin.
        '''
        log.info("Checking all methods of service [%s] for a required login..."%(service,))

        for fn in dir(service):

            f = getattr(service,fn)

            mi = xmlrpc.methodInfo(f)

            if mi is not None:
                self.assertNeedsLogin(f,*mi.getDefaultParameters())

        log.info("[OK] All methods of service [%s] require a login."%(service,))

    def performRequest(self,method,url,**kwargs):
        '''
           Perform a request on the given URL, which will be made absolute by the
           baseUrl of the test class, if it is relative.

           If a current user is set, the cookies of the LoginInfo of the current user
           will be passed to requests.reqest().

           :param method: An HTTP method like 'get', 'put' or 'post'.
           :param url: The URL under test.
           :param **kwargs: All other parameters are passed to ~requests.request without changes.
           :return: The requests library response object.
        '''
        loginInfo = getCurrentUser()

        if url[:1] == '/':
            if self.baseUrl[-1:] == '/':
                finalUrl = self.baseUrl[:-1]+url
            else:
                finalUrl = self.baseUrl+url
        else:
            finalUrl = url

        if loginInfo:
            log.info("Performing request [%s] on URL [%s] as user [%s]."%(method,url,loginInfo.principal))
            return request(method,finalUrl,cookies=loginInfo.cookies,**kwargs)
        else:
            headers = getCurrentCertificate()
            if headers:
                log.info("Performing request [%s] on URL [%s] with headers [%s]."%(method,url,headers))
                return request(method,finalUrl,headers = headers,**kwargs)
            else:
                log.info("Performing request [%s] on URL [%s] with no user."%(method,url))
                return request(method,finalUrl,**kwargs)

    def assertUrlReturnsCode(self,code,method,url,**kwargs):
        '''
           Check an URL for sending a response with the given reponse status codes.
           :param code: The expected status code like 404, 401, 200, 302.
           :param **kwargs: All other parameters are passed to ~requests.request without changes.
           :return: The requests library response object.
        '''
        response = self.performRequest(method,url,**kwargs)

        self.assertEqual(code,response.status_code,
                             "[%s] on URL [%s] did send status code [%s] instead of [%s]."%(method,url,response.status_code,code))
        log.info("[OK] [%s] on URL [%s] sent status [%s]."%(method,response.url,response.status_code))
        return response

    def assertUrlSucceeds(self,method,url,**kwargs):
        '''
           Check an URL for sending a 200 response
           :param method: An HTTP method like 'get', 'put' or 'post'.
           :param url: The URL under test.
           :param **kwargs: All other parameters are passed to ~requests.request without changes.
           :return: The requests library response object.
        '''
        return self.assertUrlReturnsCode(200,method,url,**kwargs)

    def assertUrlNeedsLogin(self,method,url,**kwargs):
        '''
           Check an URL for sending a 401 response with an X-HTTP-util-login-url
           header.
           :param method: An HTTP method like 'get', 'put' or 'post'.
           :param url: The URL under test.
           :param **kwargs: All other parameters are passed to ~requests.request without changes.
           :return: The requests library response object.
        '''
        response = self.assertUrlReturnsCode(401,method,url,**kwargs)

        loginUrl = response.headers.get("x-http-util-login-url")
        self.assertIsNotNone(loginUrl,
                             "[%s] on URL [%s] did not send an [X-HTTP-util-login-url] header in a 401 response."%(method,url))
        log.info("[OK] [%s] on URL [%s] requires a login from [%s]."%(method,url,loginUrl))
        return response

    def assertUrlUnauthorized(self,method,url,**kwargs):
        '''
           Check an URL for sending a 401 response
           :param method: An HTTP method like 'get', 'put' or 'post'.
           :param url: The URL under test.
           :param **kwargs: All other parameters are passed to ~requests.request without changes.
           :return: The requests library response object.
        '''
        return self.assertUrlReturnsCode(401,method,url,**kwargs)

    def assertUrlForbidden(self,method,url,**kwargs):
        '''
           Check an URL for sending a 403 response
           :param method: An HTTP method like 'get', 'put' or 'post'.
           :param url: The URL under test.
           :param **kwargs: All other parameters are passed to ~requests.request without changes.
           :return: The requests library response object.
        '''
        return self.assertUrlReturnsCode(403,method,url,**kwargs)
