import os.path
import io
import json
from typing import Sequence
import requests
import urllib.parse
import datetime
from dataclasses import dataclass
import typing as t
import traceback
import logging
from contextlib import AbstractContextManager
from types import TracebackType

log = logging.getLogger(__name__)


def formatTimespan(dt: datetime.timedelta)->str:

    if dt.days < 0:
        raise ValueError("Cannot serialize negative timespan.")

    hours = dt.seconds // 3600
    minutes = (dt.seconds // 60) % 60
    seconds = 1 if dt.days == 0 and dt.seconds == 0 else dt.seconds % 60

    with io.StringIO() as out:
        if dt.days > 0:
            out.write(str(dt.days))
            out.write('d')

        if hours > 0:
            if out.tell() > 0: out.write(' ')
            out.write(str(hours))
            out.write('h')

        if minutes > 0:
            if out.tell() > 0: out.write(' ')
            out.write(str(minutes))
            out.write('m')

        if seconds > 0:
            if out.tell() > 0: out.write(' ')
            out.write(str(seconds))
            out.write('s')

        ret = out.getvalue()
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug(f"formatTimespan({dt})={ret}")
        return ret

STATUS_Passed   = 1
STATUS_Blocked  = 2
STATUS_Untested = 3
STATUS_Retest   = 4
STATUS_Failed   = 5

class TestrailStepResult:
    """
        A testrail step result.
        See: https://support.gurock.com/hc/en-us/articles/7101746388500#configuring-separate-test-steps-0-0
    """
    def __init__(self,
                status_id:int,
                content:str,
                expected:t.Optional[str] = None,
                actual:t.Optional[str] = None):
        """_summary_

        :param status_id: The status, which is STATUS_Passed, STATUS_Blocked, ..., STATUS_Failed
        :type status_id: int
        :param content: Identifier for this step. Could just be numeric like 'Step 1' or more descriptive
            like 'Test API endpoint authentication'
        :type content: str
        :param expected: Describe what the expected outcome was
        :type expected: str
        :param actual: Describe what the actual outcome was. Could be omitted on 'STATUS_Passed'.
        :type actual: str
        """
        self.status_id      = status_id
        self.content        = content
        self.expected       = expected
        self.actual         = actual

    def __str__(self) -> str:
        return f"TestrailStepResult[status_id={self.status_id},content={self.content},expected={self.expected},actual={self.actual}]"


class TestrailResult:
    """
       A testrail result to POST as tesrun result.
    """
    def __init__(self,
                 status_id:int,
                 comment:str,
                 version:t.Optional[str] = None,
                 elapsed:t.Optional[datetime.timedelta] = None,
                 defects:t.Optional[str] = None,
                 assignedto_id:t.Optional[int] = None,
                 step_results: t.Optional[Sequence[TestrailStepResult]] = None) -> None:
        """_summary_

        :param status_id: The status, which is STATUS_Passed, STATUS_Blocked, ..., STATUS_Failed
        :type status_id: int
        :param comment: A comment for the result to show in testrail
        :type comment: str
        :param version: The application version tested, defaults to None
        :type version: str, optional
        :param elapsed: The time elapsed for the test, defaults to None
        :type elapsed: datetime.timedelta, optional
        :param defects: A comma-separated list of defct IDs, defaults to None
        :type defects: str, optional
        :param assignedto_id: The ID of hte user the case will be assigned to, defaults to None
        :type assignedto_id: int, optional
        :param step_results: An optional list of step results for this case. If provided, they will
            be included in the API call in the 'custom_step_results' field. For more information, see
            https://support.gurock.com/hc/en-us/articles/7077819312404-Results#addresult
        """
        self.status_id     = status_id
        self.comment       = comment
        self.version       = version
        self.elapsed       = elapsed
        self.defects       = defects
        self.assignedto_id = assignedto_id
        self.step_results  = step_results

    def __str__(self) -> str:
        return f"TestrailResult[status_id={self.status_id},comment={self.comment},version={self.version},elapsed=[{self.elapsed}],defects=[{self.defects}],assignedto_id={self.assignedto_id},step_results=[{self.step_results}]]"

class TestrailEncoder(json.JSONEncoder):
    def default(self, d:object)->object:
        if isinstance(d,datetime.timedelta):
            return formatTimespan(d)
        if isinstance(d,TestrailResult):
            return {("custom_step_results" if key == "step_results" else key):value for (key, value) in d.__dict__.items()}
        if isinstance(d,TestrailStepResult):
            return d.__dict__
        else:
            return super().default(d)

class TestrailServer:
    """
    A testrail server as read with authentication credentials from
    ~/.config/clazzes/testrail/servers.json

    :ivar serverId: The ID of the server used on command line arguments, mostly
                    a string like 'clazzes' or 'iteg'.
    :ivar baseUrl: The base URL, which is a schema and a server name like
                   'https://testrail.clazzes.org'.
    :ivar user: The user ID for authenticating against testrail.
    :ivar apiKey: The API key for the given user for authenticating against testrail.
    """

    __slots__ = ("serverId","baseUrl","user","apiKey")

    def __init__(self,serverId:str,baseUrl:str,user:str,apiKey:str) -> None:
        self.serverId = serverId
        self.baseUrl = baseUrl
        self.user = user
        self.apiKey = apiKey

    def performRequest(self,*parts:object, payload:object=None) -> requests.Response:
        """
           Perform an API request by concatenating all positional arguments with
           a slash.

        :param payload: The payload for POST requests, which is JSON-encoded, defaults to None
        :return: A response object with a postive HTTP status code.
        :rtype: requests.Repsonse
        """

        url = self.baseUrl

        if not url.endswith('/'):
            url += '/'

        url += "index.php?/api/v2"

        for part in parts:
            url += "/"
            if isinstance(part,str):
                url += urllib.parse.quote(part,safe=())
            else:
                url += urllib.parse.quote(str(part),safe=())

        log.info(f"Performing request to [{url}]")

        if payload is None:
            resp = requests.get(url,auth=(self.user,self.apiKey),
                                headers={"accept":"application/json"})
        else:

            data = json.dumps(payload,cls=TestrailEncoder)

            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug(f"POSTing data [{data}]")

            resp = requests.post(url,
                                 auth=(self.user,self.apiKey),
                                 data=data,
                                 headers={"content-type":"application/json",
                                          "accept":"application/json"
                                 })

        log.info(f"Request to [{url}] returned [{resp.status_code}]")

        if 400 <= resp.status_code < 500 or 500 <= resp.status_code < 600:
            log.error(f"Body of request to [{url}] was:\n{resp.text}")

        resp.raise_for_status()
        return resp

    def add_result_for_case(self,run_id:int, case_id:int, result: TestrailResult) -> requests.Response:
        """

            POST a testrail result to /api/v2/add_result_for_case/{run_id}/{case_id}

        :param run_id: The testrail run ID.
        :type run_id: int
        :param case_id: The testrail case ID.
        :type case_id: int
        :param result: The testrail result to POST.
        :type result: TestrailResult
        :return: A response object with a postive HTTP status code.
        :rtype: requests.Response
        """
        return self.performRequest("add_result_for_case",run_id,case_id,
                                   payload=result)

def readServers(filename : str)->t.List[TestrailServer]:

    log.info(f"Loading list of testrail servers from file [{filename}].")

    with open(filename,"r",encoding='utf-8') as fh:
        loaded = json.load(fh,object_hook=lambda x : TestrailServer(**x))
        if isinstance(loaded, list):
            return loaded
        else:
            raise TypeError("Json isn't a list.")

def readUserServers()->t.List[TestrailServer]:

    fn = os.path.expanduser('~/.config/clazzes/testrail/servers.json')

    if os.path.exists(fn):
        return readServers(fn)
    else:
        log.warning(f"File [{fn}] does not exists, add a content like follows to it:")
        log.warning("[")
        log.warning('  { "serverId": "clazzes",')
        log.warning('    "baseUrl": "https://testrail.clazzes.org/",')
        log.warning('    "user": "myuser",')
        log.warning('    "apiKey": "<api-key>"')
        log.warning('  }')
        log.warning("]")
        return []

def getServer(servers:Sequence[TestrailServer],serverId:str) -> t.Optional[TestrailServer]:
    for server in servers:
        if server.serverId == serverId:
            log.info(f"Server with ID [{serverId}] has URL [{server.baseUrl}] and user [{server.user}]")
            return server

    log.warning(f"Cannot find server with ID [{serverId}] in list of servers.")
    return None

def getUserServer(serverId:str) -> t.Optional[TestrailServer]:
    return getServer(readUserServers(),serverId)


class TestrailCase(AbstractContextManager):

    def __init__(self, server: t.Optional[TestrailServer], runId: t.Optional[int], caseId: int, version: t.Optional[str] = None, reentrant: bool=False):
        """

           Initialize the testrail context manager.

        :param server: The testrail server to communicate with. If None, testrail reporting is disabled.
        :type server: TestrailServer
        :param runId: The ID of the test run on testrail. If None, testrail reporting is disabled.
        :type runId: int
        :param caseId: The ID of the test case on testrail.
        :type caseId: int
        :param version: The optional application version
        :type version: str
        :param reentrant: If True, test errors are not rethrown because this test may not have any impact
                    on follow-up tests.
        :type reentrant: bool
        """
        self.server = server
        self.runId = runId
        self.caseId = caseId
        self.version = version
        self.reentrant = reentrant
        self.start = datetime.datetime.now()
        self.end: t.Optional[datetime.datetime] = None
        self.testrailStatus: t.Optional[bool] = None
        self.stepResults: list[TestrailStepResult] = []

    def commit(self, result: TestrailResult)->None:
        # set the committed flag in order to not propagate testrail errors
        # arising below to testrail itself.
        self.end = datetime.datetime.now()

        result.elapsed = self.end - self.start

        # set the status on the server.
        if self.server is not None and self.runId is not None:
            log.info(f"Propagating testrail result [{result}] for case [{self.caseId}]")
            try:
                self.server.add_result_for_case(self.runId,self.caseId,result)
                self.testrailStatus = True
            except:
                self.testrailStatus = False
                raise
        else:
            log.warning(f"Not propagating testrail result [{result}] for case [{self.caseId}]")

    def add_step_result(self, status_id:int, content:str, expected:t.Optional[str] = None, actual:t.Optional[str] = None)->None:
        self.stepResults.append(TestrailStepResult(status_id, content, expected, actual))

    def passed(self,comment:str)->None:
        result = TestrailResult(status_id=STATUS_Passed,
                                comment=comment,
                                version=self.version,
                                step_results=self.stepResults)
        self.commit(result)

    def failed(self,comment:str)->None:
        result = TestrailResult(status_id=STATUS_Failed,
                                comment=comment,
                                version=self.version,
                                step_results=self.stepResults)
        self.commit(result)

    def __enter__(self)->"TestrailCase":
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> bool:

        # If testrailStatus is set to False, then the exception was caught when trying to connect to TestRail.
        # Such exceptions are propagated "as is" by returning False (causes exceptions to be re-thrown)
        if self.testrailStatus == False:
            return False

        if self.end is None:

            if exc_type is None and exc_val is None and exc_tb is None:
               raise ValueError(f"Test case [{self.caseId}] has not been committed, please fix your code and add tc.passed() or tc.failed().")

            comment = "".join(traceback.format_exception(exc_type, exc_val,
                                          exc_tb))
            try:
                self.failed(comment)
            except:
                log.exception("Error propagating execption to testrail")
                return False

        return self.reentrant

    def step(self, content: str, expected: str)->"TestrailStep":
        return TestrailStep(content,expected,self)

class TestrailStep(AbstractContextManager):
    tc: TestrailCase
    content:str
    expected:str
    committed: bool
    status_id: int
    def __init__(self, content: str, expected: str, tc: TestrailCase):
        """
            Initialize a Testrail Step context manager.

        :param content: A short description of this test case. Could also just be numerical, like 'Step 1'.
        :type content: str
        :param expected: Description of the expected output
        :type expected: str
        """
        self.tc = tc
        self.content = content
        self.expected = expected
        self.committed = False
        self.status_id = 3

    def commit(self, status_id: int, actual:t.Optional[str])->None:
        self.committed = True
        self.tc.add_step_result(status_id, self.content, self.expected, actual)

    def passed(self, actual:t.Optional[str] = None)->None:
        self.commit(STATUS_Passed, actual)

    def failed(self, actual:t.Optional[str] = None)->None:
        self.commit(STATUS_Failed, actual)

    def __enter__(self)->"TestrailStep":
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> t.Literal[False]:

        if not self.committed:
            if exc_type is None and exc_val is None and exc_tb is None:
                # In this case, the step was finished without exceptions, but without using passed() first -> remind user
                # log.warning(f"Step [{self.content}] of Test case [{self.tc.caseId}] has not been committed, please fix your code and add ts.passed() or ts.failed().")
                raise ValueError(f"Step [{self.content}] of Test case [{self.tc.caseId}] has not been committed, please fix your code and add ts.passed() or ts.failed().")
            else:
                # In this case, an Exception(likely AssertionError) was thrown within the step context, so we commit a failure
                self.actual = f"{exc_val}"

                self.failed(self.actual)
                return False

        #If step has already been commited without issue, exit cleanly
        return False

@dataclass
class TestrailConfig:
    testrail_server: t.Optional[TestrailServer]
    testrail_run_id: t.Optional[int]
    application_version: t.Optional[str] = None

    def case(self, case_id: int, reentrant: bool = False)->TestrailCase:
        return TestrailCase(self.testrail_server,self.testrail_run_id,case_id,self.application_version,reentrant=reentrant)
