'''
..
  $Id: proxy.py 97 2018-08-23 12:53:55Z pytools $
  
  OSGi python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 09.04.2020
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import sys
import logging
import getopt
import json
from unittest import TestResult
from paramiko.rsakey import RSAKey
from clazzes.osgitools.ssh import OSGiClient
from clazzes.osgitools.testrail import TestrailServer,TestrailResult,getUserServer

from clazzes.tools.datetools import parse_duration

log = logging.getLogger(__name__)

def usage(argv0):
    print("Usage: %s [options...] [<cmd1> [<cmd> ...]]"%argv0,file=sys.stderr)
    print("  Supported options are:",file=sys.stderr)
    print("  -p <port> SSH port to use. Default is 8101.",file=sys.stderr)
    print("  -k <pkey> SSH private key file used to authenticate.",file=sys.stderr)
    print("            Default is to use the SSH agent for private keys.",file=sys.stderr)
    print("  -K <khf>  SSH known_hosts file used to check the host key.",file=sys.stderr)
    print("            Default is to use the system default known_hosts file.",file=sys.stderr)
    print("  -d <name> The name of the JDBC datasource to use for DB commands.",file=sys.stderr)
    print("  -t <tables> A comma-separated list of database tables to delete",file=sys.stderr)
    print("              for the restartDB command.",file=sys.stderr)
    print("  -s <serverId> The testrail server to contact from the list in",file=sys.stderr)
    print("              ~/.config/clazzes/testrail/servers.json.",file=sys.stderr)
    print("  -r <runId> The testrail run ID for POSTing results.",file=sys.stderr)
    print("  -c <caseId> The testrail case ID for POSTing results.",file=sys.stderr)
    print("  -b <names> A comma-separated list of bundle names to resolve",file=sys.stderr)
    print("              or restart for the restartDB command.",file=sys.stderr)
    print("  -h  Print this help text.",file=sys.stderr)
    print("  Supported commands are:",file=sys.stderr)
    print("    findBundleId  ...Resolve the given bundle names to bundle IDs.",file=sys.stderr)
    print("    startBundle   ...Start the bundles given by their names.",file=sys.stderr)
    print("    stopBundle    ...Stop the bundles given by their names.",file=sys.stderr)
    print("    restartDB     ...Stop the bundles given by their names,",file=sys.stderr)
    print("                     delete given DB tables and restart bundles.",file=sys.stderr)
    print("    countDBTables ...Count the DB tables in the given datasource.",file=sys.stderr)
    print("    testrailApi   ...Perform a GET query to the testrail API.",file=sys.stderr)
    print("    testrailResult...POST a testrail result via the API.",file=sys.stderr)

def pretty_print(obj):
    return json.dumps(obj,sort_keys=True,indent=2)

def parse_result(cmdline) -> TestrailResult:
    
    kwargs = {}

    for x in cmdline:
        i = x.index("=")
        k = x[:i]
        v = x[i+1:]
        if k == "elapsed":
            kwargs[k] = parse_duration(v)
        elif k in ("status_id","assignedto_id"):
            kwargs[k] = int(v)
        else:
            kwargs[k] = v

    return TestrailResult(**kwargs)

def run_tool(argv0,args):
    '''
      The main program for accessing osgi-runner via ssh
      :param argv0: The name under which the program has been invoked.
      :param args: Command line arguments 
    '''
    try:
        opts, commands = getopt.getopt(args, "p:k:K:d:t:b:s:c:r:h", ["port=","key=","known_hosts=","datasource=","tables=","bundles=","server=","case=","run=","help"])
    except getopt.GetoptError as err:
        # print help information and exit:
        print(err,file=sys.stderr)  # will print something like "option -a not recognized"
        usage(sys.argv[0])
        sys.exit(64)
    
    ssh_port=8101
    ssh_key = None
    ssh_known_hosts = None
    datasource = None
    serverId = None
    caseId = None
    runId = None
    tables = []
    bundles = []
        
    for o, a in opts:
        if o in ("-h,--help"):
            usage(sys.argv[0])
            sys.exit(0)
        elif o in ("-k", "--key"):
            ssh_key=RSAKey.from_private_key_file(a)
        elif o in ("-K", "--known_hosts"):
            ssh_known_hosts = a
        elif o in ("-p", "--port"):
            ssh_port=int(a)
        elif o in ("-d", "--datasource"):
            datasource=a
        elif o in ("-t", "--tables"):
            tables.extend(a.split(","))
        elif o in ("-b", "--bundles"):
            bundles.extend(a.split(","))
        elif o in ("-s", "--server"):
            serverId=a
        elif o in ("-c", "--case"):
            caseId=int(a)
        elif o in ("-r", "--run"):
            runId=int(a)
        else:
            assert False, "unhandled option"
    
    if len(commands) == 0:
        log.warning("No commands given.")
        usage(sys.argv[0])
        sys.exit(64)

    if commands[0] == "testrailApi":
        
        tr = getUserServer(serverId)
        resp = tr.performRequest(*commands[1:])
        
        res = resp.json()
        log.info(f"Testrail API returned [{pretty_print(res)}]")

    elif commands[0] == "testrailResult":

        tr = getUserServer(serverId)

        result = parse_result(commands[1:])
        
        log.info(f"POSTing {result}...")

        resp=tr.add_result_for_case(runId,caseId,result)

        res = resp.json()
        log.info(f"Testrail add_result_for_case returned [{pretty_print(res)}]")

    else:

        with OSGiClient(ssh_port,ssh_key,ssh_known_hosts,datasource) as client:
            client.connect()
        
            for cmd in commands:
                log.info("Executing command [%s]"%cmd)
                if cmd == "findBundleId":
                    ret = client.findBundleIds(bundles)
                    for b in ret:
                        if b[0] is None:
                            log.info("Cannot bundle ID for name [%s]."%b[1])
                        else:
                            log.info("Found bundle ID [%d] for name [%s]."%b)
                elif cmd == "stopBundle":
                    ret = client.findBundleIds(bundles,True)
                    for b in ret:
                        log.info("Stopping bundle [%s/%s]."%b)
                        client.stopBundle(*b)
                elif cmd == "startBundle":
                    ret = client.findBundleIds(bundles,True)
                    for b in ret:
                        log.info("Starting bundle [%s/%s]."%b)
                        client.startBundle(*b)
                elif cmd == "restartDB":
                    client.restartDB(bundles,tables)
                elif cmd == "countDBTables":
                    ret = client.countDBTables()
                    log.info("Number of DB tables is [%s]."%ret)
                else:
                    log.warn("Unknown command [%s]"%cmd)