'''
..
  $Id: proxy.py 97 2018-08-23 12:53:55Z pytools $

  OSGi python tools of the clazzes.org project
  http://www.clazzes.org

  Created: 07.04.2020

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import logging
from time import sleep
import typing as t
import sys
from unittest import TestCase
from contextlib import AbstractContextManager

T = t.TypeVar("T")

# enter*Context for python versions under 3.11
if sys.version_info >= (3, 11):
    from unittest import enterModuleContext

    def enterContext(tc: TestCase, cm: AbstractContextManager[T]) -> T:
        return tc.enterContext(cm)

    def enterClassContext(cls: t.Type[TestCase], cm: AbstractContextManager[T]) -> T:
        return cls.enterClassContext(cm)
else:
    from unittest import addModuleCleanup

    def _enter_context(cm, addcleanup):
        # We look up the special methods on the type to match the with
        # statement.
        cls = type(cm)
        try:
            enter = cls.__enter__
            exit = cls.__exit__
        except AttributeError:
            raise TypeError(f"'{cls.__module__}.{cls.__qualname__}' object does "
                            f"not support the context manager protocol") from None
        result = enter(cm)
        addcleanup(exit, cm, None, None, None)
        return result

    def enterModuleContext(cm: AbstractContextManager[T]) -> T:
        return _enter_context(cm,addModuleCleanup)

    def enterContext(tc: TestCase, cm: AbstractContextManager[T]) -> T:
        return _enter_context(cm,tc.addCleanup)

    def enterClassContext(cls: t.Type[TestCase], cm: AbstractContextManager[T]) -> T:
        return _enter_context(cm,cls.addClassCleanup)



log = logging.getLogger(__name__)



R = t.TypeVar("R")


def retry_operation_simple(
    operation: str,
    func: t.Callable[[], R],
    ntries: int = 10,
    interval: float = 1,
    assert_return: t.Optional[R] = None,
) -> R:
    return retryOperation(
        operation,
        func,
        ntries=ntries,
        interval=interval,
        assert_return=assert_return,
    )

def retryOperation(operation: str,
                   func : t.Callable[...,R],
                   args:t.Sequence[t.Any]=(),
                   kwargs:t.Mapping[t.Any,t.Any]={},
                   ntries: int = 10,
                   interval:float = 1,
                   assert_return: t.Optional[R] = None)->R:
    '''
        Perform an operation and retry up to ntries times.
        :param operation: A string used for logging the operation.
        :param func: The function to execute.
        :param args: Optional positional parameters to pass to func.
        :param kwargs: Optional named parameters to pass to func.
        :param ntries: Number of tries
        :param assert_return: If not None, run until the given value is returned.
    '''
    for i in range(1,ntries):
        try:
            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("Performing operation [%s], try [%d/%d]..."%(operation,i,ntries))

            ret = func(*args,**kwargs)

            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("Operation [%s] returned [%s] in try [%d/%d]."%(operation,ret,i,ntries))

            if assert_return is None or ret == assert_return:
                return ret

            log.warning("Operation [%s] returned [%s] instead of [%s] in try [%d], will retry in [%ds]..."%
                        (operation,ret,assert_return,i,interval))

            sleep(interval)

        except Exception:
            log.warning("Operation [%s] failed in try [%d], will retry in [%ds]..."%
                        (operation,i,interval))
            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug("Exception of operation [%s] was:"%(operation,),exc_info=True)
            sleep(interval)

    if log.getEffectiveLevel() <= logging.DEBUG:
        log.debug("Performing operation [%s], final try [%d]..."%(operation,ntries))

    ret = func(*args,**kwargs)

    if log.getEffectiveLevel() <= logging.DEBUG:
        log.debug("Operation [%s] returned [%s] in final try [%d]."%(operation,ret,ntries))

    if assert_return is None or ret == assert_return:
        return ret

    raise ValueError("Operation [%s] returned [%s] instead of [%s] in final try [%d]."%
                        (operation,ret,assert_return,ntries))
